/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "fileio.h"
#include "frap.h"
#ifdef ZLIB
#include <zlib.h>
#endif

static char **kn = NULL, **know = NULL;
static int **im = NULL;
static int savecnt = 1;

typedef struct { int x,y; } point;

#define STACKSIZE 100000

point *stack = NULL;

int stacktop;

static void sinit()
{
  stacktop = 0;
}

static int sempty()
{
  return !stacktop;
}

static void spop(int *x, int *y)
{
  if(stacktop) 
    stacktop--;
  else
    fprintf(stderr,"INTERNAL ERROR: Stack underflow!\n");
  *x = stack[stacktop].x;
  *y = stack[stacktop].y;
}

static void spush(int x, int y)
{
  stack[stacktop].x = x;
  stack[stacktop].y = y;
  kn[y][x] |= 16;
  assert(stacktop++ < STACKSIZE);
}

static void check(int x, int y)
{
  know[y][x] = 15;
  know[(y+1)%height][x] |= 1;
  know[y][(x+width-1)%width] |= 2;
  know[(y+height-1)%height][x] |= 4;
  know[y][(x+1)%width] |= 8;
}

typedef struct 
{
  int n,freq;
} freqT;

static int compare1(const void *v1, const void *v2)
{
  freqT *a = (freqT *) v1;
  freqT *b = (freqT *) v2;
  if(a->freq != b->freq) return b->freq - a->freq;
  return a->n - b->n;
}

static int compare2(const void *v1, const void *v2)
{
  freqT *a = (freqT *) v1;
  freqT *b = (freqT *) v2;
  return a->n - b->n;
}

static void output_code(char *c, int clen, int *freq, FILE *out,
			int destlen, unsigned char *dest)
{
  freqT f[8];
  int i,j,fq[7];
  double entr,p;

  freq[0] = -1;
  for(i=0;i<8;i++) { f[i].n = i; f[i].freq = freq[i]; }
  /* pass 1 */
  qsort(f,8,sizeof(freqT),compare1);
  qsort(f,4,sizeof(freqT),compare2);
  for(i=0;i<4;i++) { f[i].freq += freq[i+8]; freq[f[i].n] += freq[i+8]; }
  for(i=0;i<clen;i++)
    if(c[i]/8 == 1) c[i] = f[c[i]%8].n;
  /* pass 2 */
  qsort(f,4,sizeof(freqT),compare1);
  qsort(f,2,sizeof(freqT),compare2);
  for(i=0;i<2;i++) { f[i].freq += freq[i+16]; freq[f[i].n] += freq[i+16]; }
  for(i=0;i<clen;i++)
    if(c[i]/8 == 2) c[i] = f[c[i]%8].n;

  /* output frequecies */
  entr = 0.0;
  for(i=0;i<7;i++)
    if(f[i].freq)
      {
	p = (double)f[i].freq/(double)clen;
	entr -= p*log(p)/log(2.0);
      }

#ifdef ZLIB
  if(entr*clen/8.0 > destlen)
    {
      /* Use zlib compression */
      pack(1,1,out);
      if(packpos%8)
	pack(0,8-packpos%8,out); /* flush buffer to next byte */
      fwrite(dest,1,destlen,out);
      return;
    }
#endif
  /* Use arithmetic coding */
  pack(0,1,out);
  for(i=0;i<7;i++)
    for(j=0;j<7;j++)
      if(f[j].n == i+1)
	{
	  pack(f[j].freq,(int)ceil(log(width*height)/log(2.0)),out);
	  fq[i] = f[j].freq;
	  break;
	}

  for(i=0;i<clen;i++) c[i]--;
  arith_init(1,out);
  arith_code(c,clen,fq);
}

void output(char *fname)
{
  int i,j,x,y,dir,m;
  int bit[50],code,b,cnt,clen;
  static char *buffer;
#ifdef ZLIB
  long unsigned int destlen,zbits = 0;
  static unsigned char *zbuffer,*dest;
#endif
  int um,vm,dombits;
  int start[100][3],startnum=0;
  FILE *out;
  rangeT *r;

  out = fopen(fname,"wb");
  fprintf(out,"FRAP0"); /* output program identifier */
  if(kn == NULL)
    {
      im = (int **) malloc(height*sizeof(int *));
      im[0] = (int *) malloc(width*height*sizeof(int));
      kn = (char **) malloc(height*sizeof(char *));
      kn[0] = (char *) malloc(width*height*sizeof(char));
      know = (char **) malloc(height*sizeof(char *));
      know[0] = (char *) malloc(width*height*sizeof(char));
      stack = (point *) malloc(STACKSIZE*sizeof(point));
      for(i=1;i<height;i++) 
	{
	  im[i] = width+im[i-1];
	  kn[i] = width+kn[i-1];
	  know[i] = width+know[i-1];
	}
      buffer = (char *) malloc(width*height);
#ifdef ZLIB
      zbuffer = (unsigned char *) calloc(2*width*height,1);
      dest = (unsigned char *) malloc(2*width*height);
#endif
    }

  /* 1. Output general information on the image */

  packpos = 0;
  pack(width,12,out);
  pack(height,12,out);
  pack(depth==3?1:0,1,out);
  pack(basesize,8,out);
  pack(s_bits,4,out);
  pack(o_bits,4,out);
  pack(rangenum,(int)ceil(log(width*height)/log(2.0)),out);

  /* 2. Output range transformation information */

  dombits = (int)ceil(log(width*height)/log(2.0)-2.0);

  cnt = 0;
  for(i=0;i<width*height;i++)
    if(rptr[i]->flag != savecnt)
      {
	r = rptr[i];
	r->flag = savecnt;

	pack(r->code[0].qscaling,s_bits,out);
	pack(r->code[0].qoffset,o_bits,out);
	pack(r->code[0].rot,3,out);
	pack(r->code[0].domain,dombits,out);

	for(j=0;j<r->size;j++)
	  im[r->pos[j]/width][r->pos[j]%width] = cnt;
	cnt++;
      }
  savecnt++;

  /* 3. Output color information (if necessary) */

  if(image->depth == 3)
    {
#ifdef ZLIB
      pack(1,1,out); /* use zlib compression for color data */
#else      
      pack(0,1,out); /* use no compression for the color data */
#endif
      for(i=j=0;i<width*height;i++)
	if(rptr[i]->flag != savecnt)
	  {
	    r = rptr[i];
	    r->flag = savecnt;
	    um = (int)rint(r->umean/(double)(basesize*basesize*r->size));
	    vm = (int)rint(r->vmean/(double)(basesize*basesize*r->size));
	    if(um < -127) um = -127;
	    if(um > 128) um = 128;
	    if(vm < -127) vm = -127;
	    if(vm > 128) vm = 128;
#ifdef ZLIB
	    zbuffer[j++] = um+127;
	    zbuffer[j++] = vm+127;
#else
	    pack(um+127,8,out);
	    pack(vm+127,8,out);
#endif
	  }
      savecnt++;
#ifdef ZLIB
      destlen = 2*width*height;
      compress(dest,&destlen,zbuffer,j);
      pack(destlen,(int)ceil(log(2*width*height)/log(2.0)),out);
      if(packpos%8)
	pack(0,8-packpos%8,out); /* flush buffer to next byte */
      fwrite(dest,1,destlen,out);
#endif
    }

  /* 4. Output partition information */

  for(y=0;y<height;y++)
    for(x=0;x<width;x++)
      {
	kn[y][x] = 
	  1*(im[y][x] != im[y][(x+1)%width]) +
	  2*(im[y][(x+1)%width] != im[(y+1)%height][(x+1)%width]) +
	  4*(im[(y+1)%height][(x+1)%width] != im[(y+1)%height][x]) +
	  8*(im[(y+1)%height][x] != im[y][x]);
	know[y][x] = 0;
      }

  clen = 0;
  for(i=0;i<50;i++) bit[i] = 0;

  sinit();

  while(1)
    {
      /* find connected component that has not been described yet */
      for(y=0;y<height;y++)
	for(x=0;x<width;x++) if(kn[y][x] & 15) goto start;
      break;
    start:
      if(kn[y][x] & 16) fprintf(stderr,"INTERNAL ERROR #1!\n");
      /* output start position */
      start[startnum][0] = x;
      start[startnum][1] = y;
      start[startnum][2] = kn[y][x];
      startnum++;
      check(x,y);
      spush(x,y);
      while(!sempty())
	{
	  while(!sempty()) { spop(&x,&y); if(kn[y][x] & 15) break; }
	  if((kn[y][x] & 15) == 0) break;
	  dir = 0;
	  while(1)
	    {
	      while(0 == (kn[y][x] & 1<<dir)) dir = (dir+1)%4;
	      kn[y][x] -= 1<<dir; /* do not use that direction again */
	      if(kn[y][x] & 15) spush(x,y);
	      switch(dir)
		{
		case 0: y = (y+height-1)%height; break;
		case 1: x = (x+1)%width; break;
		case 2: y = (y+1)%height; break;
		case 3: x = (x+width-1)%width; break;
		}
	      kn[y][x] -= 1<<((dir+2)%4);
	      if(kn[y][x] & 16) break;
	      code = 0;
	      for(b=m=0;m<4;m++)
		{
		  if(!(know[y][x] & 1<<(dir+3+m)%4))
		    {
#ifdef ZLIB
		      if(zbits%8 == 0) zbuffer[zbits/8] = 0;
#endif
		      if(kn[y][x] & 1<<(dir+3+m)%4)
			{
			  code |= 1<<b;
#ifdef ZLIB
			  zbuffer[zbits/8] |= (1<<(zbits%8));
#endif
			}
#ifdef ZLIB
		      zbits++;
#endif
		      b++;
		    }
		}
	      if(!b) continue;
	      if(b==3&&!code) fprintf(stderr,"INTERNAL ERROR #2!\n");
	      if(b>3) fprintf(stderr,"INTERNAL ERROR #3!\n");
	      code += 8*(3-b);
	      bit[code]++;
	      buffer[clen++] = (char) code;
	      check(x,y);
	    }
	}
    }

  pack(startnum,(int)ceil(log(width*height)/log(2.0)),out);

  for(i=0;i<startnum;i++)
    {
      pack(start[i][0],(int)ceil(log(width)/log(2.0)),out);
      pack(start[i][1],(int)ceil(log(height)/log(2.0)),out);
      pack(start[i][2],4,out);
    }

#ifdef ZLIB
  destlen = 2*width*height;
  compress(dest,&destlen,zbuffer,(zbits+7)/8);
  output_code(buffer,clen,bit,out,destlen,dest);
  free(zbuffer); free(dest);
#else
  output_code(buffer,clen,bit,out,0,NULL);
#endif

  /* flush last byte */
  pack(0,7,out);
  fclose(out);
}
