/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "frap.h"
#include "nn_search.h"

static int timestamp = 0;
static double collerr;

int gettime()
{
  return ++timestamp;
}

/**************************************************************/
/* Join the two ranges that give the lowest increase of error */
/**************************************************************/

void merge_best()
{
  heapNodeT *b = heap_pop();
  int i;

  if(b == NULL || b->join == NULL) /* Heap empty? */
    {
      fprintf(stderr,"INTERNAL ERROR: Heap empty!\n");
      exit(1);
    }
  
  collerr += b->join->code[0].err - (*(b->r1))->code[0].err -(*(b->r2))->code[0].err;
  
  for(i=0;i<b->join->size;i++)
    {
      if(rptr[b->join->pos[i]]->size != -1)
	{
	  free(rptr[b->join->pos[i]]->pos);
	  free(rptr[b->join->pos[i]]->code);
	  rptr[b->join->pos[i]]->size = -1;
	}
      rptr[b->join->pos[i]] = b->join;
    }

  b->join->timestamp = gettime();
  rangenum--;
}

/**************************************************************/
/* Merge the two ranges r1 and r2 and return the merged range */
/**************************************************************/

rangeT *merge_ranges(rangeT *r1, rangeT *r2)
{
  int i,x,y,dx,dy,nx,ny;
  rangeT *r = (rangeT *) malloc(sizeof(rangeT)), *tmp;

  if(r2->pos[0] < r1->pos[0])
    { tmp = r1; r1 = r2; r2 = tmp; }
  r->size = r1->size + r2->size;
  r->pos = (int *) malloc(r->size*sizeof(int));
  for(i=0;i<r1->size;i++) r->pos[i] = r1->pos[i];
  for(i=0;i<r2->size;i++) r->pos[i+r1->size] = r2->pos[i];

  /* merge the two ranges */

  nx = r2->pos[0]%width - r1->pos[0]%width;
  ny = r2->pos[0]/width - r1->pos[0]/width;
  r->codenum = r1->codenum + r2->codenum;
  r->code = (codeT *) malloc(r->codenum*sizeof(codeT));
  for(i=0;i<r1->codenum;i++) r->code[i] = r1->code[i];
  for(i=0;i<r2->codenum;i++)
    {
      r->code[i+r1->codenum] = r2->code[i];
      dx = (r2->code[i].rot&2)?nx:-nx;
      dy = (r2->code[i].rot&1)?ny:-ny;
      if(r2->code[i].rot&4) /* flip */
	{ 
	  x = (r2->code[i].domain%(width2) + dy + width2*height2)%width2;
	  y = (r2->code[i].domain/(width2) + dx + width2*height2)%height2;
	}
      else
	{
	  x = (r2->code[i].domain%(width2) + dx + width2*height2)%width2;
	  y = (r2->code[i].domain/(width2) + dy + width2*height2)%height2;
	}
      r->code[i+r1->codenum].domain = y*width2+x;
      r->code[i+r1->codenum].precomp = -(r1->size);
    }

  rd_compare(r,codesize);
  r->umean = r1->umean + r2->umean;
  r->vmean = r1->vmean + r2->vmean;
  r->timestamp = gettime();
  r->flag = 0;

  return r;
}

/*************************************************/
/* Subsampling routine for downfiltering domains */
/*************************************************/

static pixel subsample(pixel **img, int x, int y)
{
  return (img[y][x] + img[y+1][x] + img[y][x+1] + img[y+1][x+1]) / 4;
}

/****************************/
/* Generate domain-codebook */
/****************************/

static void compute_codebook(pixel ***img)
{
  int i,j,x,y,dx,dy;

  for(i=y=0;y<height-1;y+=2)
    for(x=0;x<width-1;x+=2,i++)
      {
	double s = 0.0, v = 0.0;
	
	if(i) codebook[i] = codebook[i-1] + basesize*basesize;
	for(dy=j=0;dy<2*basesize;dy+=2)
	  for(dx=0;dx<2*basesize;dx+=2,j++)
	    {
	      codebook[i][j] = subsample(img[0],x*basesize+dx,y*basesize+dy);
	      s += codebook[i][j];
	      v += codebook[i][j]*codebook[i][j];
	    }
	csum[(y>>1)*width2+(x>>1)] = s;
	cvar[(y>>1)*width2+(x>>1)] = v;
      }
}

/*************************************************/
/* Initialize a new partition from image 'image' */
/*************************************************/

/* Global variables that have already been set:
   image,basesize,codesize */

void init()
{
  int dx,dy,i,j,k,l,neg,x,y;
  double *umean,*vmean;
  kdtree *kd_tree = NULL;
  pixel **nncodebook = NULL;
  pixel *nnquery = NULL;
  int *approxnn = NULL;
  double s,v,um,vm;

  /* initialize variables */

  range = NULL; baseblock = NULL; codebook = NULL; rot = NULL;
  rptr = NULL;

  width = image->width / basesize;
  height = image->height / basesize;
  depth = image->depth;
  width2 = width>>1;
  height2 = height>>1;

  /* compute rotation schemes */

  rot = (int **) malloc(8*sizeof(int *));
  rot[0] = (int *) malloc(8*basesize*basesize*sizeof(int));

  for(i=neg=0;neg<=1;neg++)
    for(dx=1;dx>=-1;dx-=2)
      for(dy=1;dy>=-1;dy-=2,i++)
	{
	  if(i) rot[i] = rot[i-1] + basesize*basesize;
	    for(j=y=0;y<basesize;y++)
	      for(x=0;x<basesize;x++,j++)
		{
		  int xx = (dx == 1)?x:basesize-x-1;
		  int yy = (dy == 1)?y:basesize-y-1;
		  if(neg) { int tmp = xx; xx = yy; yy = tmp; }
		  rot[i][j] = yy*basesize+xx;
		}
	}

  /* compute range vectors */

  baseblocknum = width*height;

  baseblock = (pixel **) malloc(baseblocknum*sizeof(pixel *));
  baseblock[0] = (pixel *) malloc(baseblocknum*basesize*basesize*sizeof(pixel));
  bsum = (double *) malloc(baseblocknum*sizeof(double));
  bvar = (double *) malloc(baseblocknum*sizeof(double));
  umean = (double *) malloc(baseblocknum*sizeof(double));
  vmean = (double *) malloc(baseblocknum*sizeof(double));
  
  for(i=y=0;y<height;y++)
    for(x=0;x<width;x++,i++)
      {
	s = 0.0, v = 0.0, um = 0.0, vm = 0.0;
	
	if(i) baseblock[i] = baseblock[i-1] + basesize*basesize;
	for(dy=j=0;dy<basesize;dy++)
	  for(dx=0;dx<basesize;dx++,j++)
	    {
	      if(depth == 3)
		{
		  um += image->img[1][y*basesize+dy][x*basesize+dx];
		  vm += image->img[2][y*basesize+dy][x*basesize+dx];
		}
	      baseblock[i][j] = image->img[0][y*basesize+dy][x*basesize+dx];
	      s += baseblock[i][j];
	      v += baseblock[i][j] * baseblock[i][j];
	    }
	bsum[i] = s;
	bvar[i] = v;
	umean[i] = um;
	vmean[i] = vm;
      }

  /* compute shrunken domain vectors */

  codebooksize = (width/2)*(height/2);

  codebook = (pixel **) malloc(codebooksize*sizeof(pixel *));
  codebook[0] = (pixel *) malloc(codebooksize*basesize*basesize*sizeof(pixel));
  csum = (double *) malloc(codebooksize*sizeof(double));
  cvar = (double *) malloc(codebooksize*sizeof(double));
  compute_codebook(image->img);

  /* initialize range partition and codes */

  nnquery = (pixel *) malloc(basesize*basesize*sizeof(pixel));
  nncodebook = (pixel **) malloc(16*codebooksize*sizeof(pixel *));
  nncodebook[0] = (pixel *) malloc(16*codebooksize*basesize*basesize*sizeof(pixel));
  for(i=k=0;i<codebooksize;i++,k+=16)
    {
      s = csum[i]/(double)(basesize*basesize);
      v = sqrt(cvar[i]-csum[i]*csum[i]/(double)(basesize*basesize));
      
      if(k)
	nncodebook[k] = nncodebook[k-1]+basesize*basesize;
      for(l=0;l<basesize*basesize;l++)
	nncodebook[k][l] = (codebook[i][l] - s)/v;
      for(j=1;j<8;j++)
	{	
	  nncodebook[k+j] = nncodebook[k]+j*basesize*basesize;
	  for(l=0;l<basesize*basesize;l++)
	    nncodebook[k+j][l] = nncodebook[k][rot[j][l]];
	}
      for(j=0;j<8;j++)
	{
	  nncodebook[k+8+j] = nncodebook[k+j]+8*basesize*basesize;
	  for(l=0;l<basesize*basesize;l++)
	    nncodebook[k+8+j][l] = -nncodebook[k+j][l];
	}
    }
  kd_tree = kdtree_build(nncodebook,codebooksize*16,basesize*basesize);
  approxnn = (int *) malloc(5*codesize*sizeof(int));

  rangenum = width*height;
  range = (rangeT *) malloc(rangenum*sizeof(rangeT));
  rptr = (rangeT **) malloc(rangenum*sizeof(rangeT *));

  collerr = 0;
  for(i=y=0;y<height;y++)
    for(x=0;x<width;x++,i++)
      {
	rptr[i] = range+i;
	range[i].size = 1;
	range[i].pos = (int *) malloc(1*sizeof(int));
	range[i].pos[0] = i;
	range[i].timestamp = 0;
	range[i].flag = 0;
	range[i].umean = umean[i];
	range[i].vmean = vmean[i];

	s = bsum[i]/(double)(basesize*basesize);
	v = sqrt(bvar[i]-bsum[i]*bsum[i]/(double)(basesize*basesize));
	
	range[i].codenum = codesize*5;
	range[i].code = (codeT *) malloc(range[i].codenum*sizeof(codeT));
	for(j=0;j<basesize*basesize;j++)
	  nnquery[j] = (baseblock[i][j] - s)/v;	    
	kdtree_search(nnquery,nncodebook,basesize*basesize,
		      kd_tree,epsilon,range[i].codenum,approxnn);
	for(j=0;j<range[i].codenum;j++)
	  {
	    range[i].code[j].domain  = approxnn[j]/16;
	    range[i].code[j].rot     = approxnn[j]%8;
	    range[i].code[j].precomp = 0;
	  }
	rd_compare(range+i,codesize); /* compute transformation parameters */
	collerr += range[i].code[0].err;
      }
  free(umean); free(vmean);
  kdtree_free(kd_tree);

  /* initialize heap */

  heap_init(2*rangenum);
  for(i=0;i<rangenum;i++)
    {
      heap_push(&rptr[i],&rptr[(i+1)%width+width*(i/width)]);
      heap_push(&rptr[i],&rptr[(i+width)%rangenum]);
    }
}

void free_all()
{
  if(baseblock != NULL)
    { free(baseblock[0]); free(baseblock); baseblock = NULL; }
  if(codebook != NULL)
    { free(codebook[0]); free(codebook); codebook = NULL; }
  if(bsum != NULL) { free(bsum); bsum = NULL; }
  if(bvar != NULL) { free(bvar); bvar = NULL; }
  if(csum != NULL) { free(csum); csum = NULL; }
  if(cvar != NULL) { free(cvar); cvar = NULL; }
  if(range != NULL) { free(range); range = NULL; }
  if(rot != NULL) { free(rot[0]); free(rot); rot = NULL; }
}
